'''
Concrete representations of coordinate transforms.

A concrete representation is typically a matrix, like a 4x4 matrix for
homogeneous coordinate vectors.

Please refer to the docs of the module `kgprim.ct.repr.mxrepr`; the other
modules are for internal use.
'''
